import math

from scipy.integrate import quad

import netto.config as config
from netto.const import correction_factor_pensions, social_security_curve


def __get_rate(salary, type, extra=0):
    return (
        social_security_curve[config.YEAR][type]["rate"] + extra
        if 0 < salary <= social_security_curve[config.YEAR][type]["limit"]
        else 0
    )


def __get_value(salary, type, extra=0):
    return min(
        salary * (social_security_curve[config.YEAR][type]["rate"] + extra),
        social_security_curve[config.YEAR][type]["limit"]
        * (social_security_curve[config.YEAR][type]["rate"] + extra),
    )


def get_rate_pension(salary):
    return __get_rate(salary, "pension")


def get_rate_unemployment(salary):
    return __get_rate(salary, "unemployment")


def get_rate_health(salary):
    extra = config.EXTRA_HEALTH_INSURANCE / 2
    return __get_rate(salary, "health", extra)


def get_rate_nursing(salary):
    extra = (
        0
        if config.HAS_CHILDREN
        else social_security_curve[config.YEAR]["nursing"]["extra"]
    )
    return __get_rate(salary, "nursing", extra)


def calc_insurance_pension(salary):
    return __get_value(salary, "pension")


def calc_insurance_unemployment(salary):
    return __get_value(salary, "unemployment")


def calc_insurance_health(salary):
    extra = config.EXTRA_HEALTH_INSURANCE / 2
    return __get_value(salary, "health", extra)


def calc_insurance_health_deductable(salary):
    extra = config.EXTRA_HEALTH_INSURANCE / 2
    return __get_value(salary, "health", extra - 0.003)


def calc_insurance_nursing(salary):
    extra = (
        0
        if config.HAS_CHILDREN
        else social_security_curve[config.YEAR]["nursing"]["extra"]
    )
    return __get_value(salary, "nursing", extra)


def calc_deductable_social_security(salary):
    return (
        math.ceil(
            calc_insurance_pension(salary) * correction_factor_pensions[config.YEAR]
        )
        + math.ceil(calc_insurance_health_deductable(salary))
        + math.ceil(calc_insurance_nursing(salary))
    )


def calc_social_security(salary):
    return round(
        calc_insurance_pension(salary)
        + calc_insurance_health(salary)
        + calc_insurance_nursing(salary)
        + calc_insurance_unemployment(salary),
        2,
    )


def calc_social_security_by_integration(salary):
    pension, _ = quad(get_rate_pension, 0, salary)
    health, _ = quad(get_rate_health, 0, salary)
    nursing, _ = quad(get_rate_nursing, 0, salary)
    unemployment, _ = quad(get_rate_unemployment, 0, salary)
    return round(pension + health + nursing + unemployment, 2)



from scipy.optimize import newton

from netto.social_security import calc_deductable_social_security, calc_social_security
from netto.taxes_income import calc_income_tax_by_integration, calc_taxable_income
from netto.taxes_other import calc_church_tax, calc_soli

"""
main.py
====================================
The core module to invoke the `calc_netto()` function
"""


def calc_netto(salary, deductable_other=0):
    """
    Return net income of a given year.
    Parameters
    ----------
    salary
        The yearly salary (float or int).
    deductable_other
        Deductables that reduce the taxable income.
    """

    deductable_social_security = calc_deductable_social_security(salary)
    taxable_income = calc_taxable_income(
        salary=salary,
        deductable_social_security=deductable_social_security,
        deductable_other=deductable_other,
    )
    income_tax = calc_income_tax_by_integration(taxable_income)
    return round(
        (
            salary
            - income_tax
            - calc_soli(income_tax)
            - calc_church_tax(income_tax)
            - calc_social_security(salary)
        ),
        2,
    )


def calc_inverse_netto(desired_netto, deductable_other=0):
    """calculate gross salary to reach desired netto"""

    def f(salary):
        return calc_netto(salary, deductable_other=deductable_other) - desired_netto

    return round(newton(f, x0=desired_netto), 0)



import unittest

import netto.config as config
import netto.social_security as social_security


class TestSocialSecurity(unittest.TestCase):
    def setUp(self):
        config.EXTRA_HEALTH_INSURANCE = 0.014
        config.CHURCH_TAX = 0.09
        config.HAS_CHILDREN = False

    def tearDown(self):
        pass

    def test_get_rate_pension(self):
        self.assertEqual(social_security.get_rate_pension(0), 0)
        self.assertEqual(social_security.get_rate_pension(10000), 0.093)
        self.assertEqual(social_security.get_rate_pension(84600), 0.093)
        self.assertEqual(social_security.get_rate_pension(84601), 0)
        self.assertEqual(social_security.get_rate_pension(100000), 0)

    def test_get_rate_health(self):
        self.assertEqual(social_security.get_rate_health(0), 0)
        self.assertEqual(social_security.get_rate_health(10000), 0.08)
        self.assertEqual(social_security.get_rate_health(58050), 0.08)
        self.assertEqual(social_security.get_rate_health(58051), 0)
        self.assertEqual(social_security.get_rate_health(100000), 0)

    def test_calc_insurance_pension(self):
        self.assertEqual(social_security.calc_insurance_pension(0), 0)
        self.assertEqual(social_security.calc_insurance_pension(10000), 0.093 * 10000)
        self.assertEqual(social_security.calc_insurance_pension(84600), 0.093 * 84600)
        self.assertEqual(social_security.calc_insurance_pension(84601), 0.093 * 84600)
        self.assertEqual(social_security.calc_insurance_pension(100000), 0.093 * 84600)

    def test_calc_insurance_health(self):
        self.assertEqual(social_security.calc_insurance_health(0), 0)
        self.assertEqual(social_security.calc_insurance_health(10000), 0.08 * 10000)
        self.assertEqual(social_security.calc_insurance_health(58050), 0.08 * 58050)
        self.assertEqual(social_security.calc_insurance_health(58051), 0.08 * 58050)
        self.assertEqual(social_security.calc_insurance_health(100000), 0.08 * 58050)

    def test_calc_deductable_social_security(self):
        self.assertEqual(social_security.calc_deductable_social_security(0), 0)
        self.assertEqual(
            social_security.calc_deductable_social_security(30000),
            2456 + 2310 + 563,  # https://www.lohn-info.de/vorsorgepauschale.html
        )

    def test_sameness_of_calc_social_security(self):
        for salary in range(0, 100001, 10000):
            self.assertEqual(
                social_security.calc_social_security(salary),
                social_security.calc_social_security_by_integration(salary),
            )

    def test_sameness_of_calc_social_security_different_config(self):
        config.EXTRA_HEALTH_INSURANCE = 0.015
        config.HAS_CHILDREN = True
        for salary in range(0, 100001, 10000):
            self.assertEqual(
                social_security.calc_social_security(salary),
                social_security.calc_social_security_by_integration(salary),
            )

    def test_get_rate_health_different_config(self):
        config.EXTRA_HEALTH_INSURANCE = 0.015
        self.assertEqual(social_security.get_rate_health(0), 0)
        self.assertAlmostEqual(social_security.get_rate_health(10000), 0.0805)
        self.assertAlmostEqual(social_security.get_rate_health(58050), 0.0805)
        self.assertEqual(social_security.get_rate_health(58051), 0)
        self.assertEqual(social_security.get_rate_health(100000), 0)

    def test_get_rate_nursing(self):
        config.HAS_CHILDREN = False
        self.assertEqual(social_security.get_rate_nursing(0), 0)
        self.assertEqual(social_security.get_rate_nursing(10000), 0.01875)
        self.assertEqual(social_security.get_rate_nursing(58050), 0.01875)
        self.assertEqual(social_security.get_rate_nursing(58051), 0)
        self.assertEqual(social_security.get_rate_nursing(100000), 0)

    def test_get_rate_nursing_different_config(self):
        config.HAS_CHILDREN = True
        self.assertEqual(social_security.get_rate_nursing(0), 0)
        self.assertEqual(social_security.get_rate_nursing(10000), 0.01525)
        self.assertEqual(social_security.get_rate_nursing(58050), 0.01525)
        self.assertEqual(social_security.get_rate_nursing(58051), 0)
        self.assertEqual(social_security.get_rate_nursing(100000), 0)



import math

from scipy.integrate import quad

import netto.config as config
from netto.const import tax_curve


def get_marginal_tax_rate(taxable_income):
    if taxable_income < tax_curve[config.YEAR][0]["step"]:
        return 0
    elif taxable_income <= tax_curve[config.YEAR][1]["step"]:
        return __calc_gradient(
            tax_curve[config.YEAR][0]["step"],
            tax_curve[config.YEAR][1]["step"],
            tax_curve[config.YEAR][0]["rate"],
            tax_curve[config.YEAR][1]["rate"],
            taxable_income,
        )
    elif taxable_income <= tax_curve[config.YEAR][2]["step"]:
        return __calc_gradient(
            tax_curve[config.YEAR][1]["step"],
            tax_curve[config.YEAR][2]["step"],
            tax_curve[config.YEAR][1]["rate"],
            tax_curve[config.YEAR][2]["rate"],
            taxable_income,
        )
    elif taxable_income < tax_curve[config.YEAR][3]["step"]:
        return tax_curve[config.YEAR][2]["rate"]
    else:
        return tax_curve[config.YEAR][3]["rate"]


def __calc_gradient(x_i, x_j, y_i, y_j, x):
    return (1 - (x_j - x) / (x_j - x_i)) * (y_j - y_i) + y_i


def calc_taxable_income(salary, deductable_social_security, deductable_other=0):
    return math.floor(
        max(0, salary - deductable_social_security - 1200 - 36 - deductable_other)
    )


def calc_income_tax(taxable_income):
    taxable_income = round(taxable_income)
    if taxable_income <= tax_curve[config.YEAR][0]["step"]:
        return 0
    elif taxable_income <= tax_curve[config.YEAR][1]["step"]:
        y = (taxable_income - tax_curve[config.YEAR][0]["step"]) / 10000
        return (
            tax_curve[config.YEAR][1]["const"][0] * y
            + tax_curve[config.YEAR][1]["const"][1]
        ) * y
    elif taxable_income <= tax_curve[config.YEAR][2]["step"]:
        z = (taxable_income - tax_curve[config.YEAR][1]["step"]) / 10000
        return (
            tax_curve[config.YEAR][2]["const"][0] * z
            + tax_curve[config.YEAR][2]["const"][1]
        ) * z + tax_curve[config.YEAR][2]["const"][2]
    elif taxable_income <= tax_curve[config.YEAR][3]["step"]:
        return (
            tax_curve[config.YEAR][2]["rate"] * taxable_income
            - tax_curve[config.YEAR][3]["const"][0]
        )
    else:
        return (
            tax_curve[config.YEAR][3]["rate"] * taxable_income
            - tax_curve[config.YEAR][3]["const"][1]
        )


def calc_income_tax_by_integration(taxable_income):
    integral, _ = quad(get_marginal_tax_rate, 0, taxable_income)
    return integral



import unittest

import netto.config as config
import netto.main as main


class TestMain(unittest.TestCase):
    def setUp(self):
        config.EXTRA_HEALTH_INSURANCE = 0.014
        config.CHURCH_TAX = 0.09
        config.HAS_CHILDREN = False

    def tearDown(self):
        pass

    def test_for_valid_main(self):
        self.assertEqual(main.calc_netto(0), 0)
        self.assertAlmostEqual(main.calc_netto(30000), 20554.38, delta=1)
        self.assertAlmostEqual(main.calc_netto(60000), 35796.68, delta=1)
        self.assertAlmostEqual(main.calc_netto(90000), 49956.92, delta=1)
        self.assertAlmostEqual(main.calc_netto(120000), 64965.08, delta=1)

    def test_for_valid_main_second_config(self):
        config.EXTRA_HEALTH_INSURANCE = 0.015
        config.CHURCH_TAX = 0.0
        config.HAS_CHILDREN = True
        self.assertAlmostEqual(main.calc_netto(30000), 20894.58, delta=1)
        self.assertAlmostEqual(main.calc_netto(60000), 36909.71, delta=1)
        self.assertAlmostEqual(main.calc_netto(90000), 52091.39, delta=1)
        self.assertAlmostEqual(main.calc_netto(120000), 68238.23, delta=1)

    def test_for_valid_inverse_netto(self):
        config.EXTRA_HEALTH_INSURANCE = 0.015
        config.CHURCH_TAX = 0.0
        config.HAS_CHILDREN = True
        self.assertEqual(main.calc_inverse_netto(main.calc_netto(10000)), 10000)
        self.assertAlmostEqual(main.calc_inverse_netto(20894.58), 30000, delta=1)
        self.assertAlmostEqual(main.calc_inverse_netto(36909.71), 60000, delta=1)
        self.assertAlmostEqual(main.calc_inverse_netto(52091.39), 90000, delta=1)
        self.assertAlmostEqual(main.calc_inverse_netto(68238.23), 120000, delta=1)



import netto.config as config
from netto.const import soli_curve


def calc_soli(tax_assessment):
    return round(
        max(
            min(
                max(0, tax_assessment - soli_curve[config.YEAR]["start_taxable_income"])
                * soli_curve[config.YEAR]["start_fraction"],
                tax_assessment * soli_curve[config.YEAR]["end_rate"],
            ),
            0,
        ),
        2,
    )


def calc_church_tax(tax_assessment):
    return round(max(tax_assessment * config.CHURCH_TAX, 0), 2)



import unittest

import netto.config as config
import netto.taxes_other as taxes_other


class TestTaxesOther(unittest.TestCase):
    def setUp(self):
        config.EXTRA_HEALTH_INSURANCE = 0.014
        config.CHURCH_TAX = 0.09
        config.HAS_CHILDREN = False

    def tearDown(self):
        pass

    def test_valid_calc_soli(self):
        self.assertEqual(taxes_other.calc_soli(-1000), 0)
        self.assertEqual(taxes_other.calc_soli(0), 0)
        self.assertEqual(taxes_other.calc_soli(16956), 0)
        self.assertAlmostEqual(taxes_other.calc_soli(16957), 0.119, delta=0.1)
        self.assertAlmostEqual(taxes_other.calc_soli(17514.96), 66.48, delta=0.1)
        self.assertAlmostEqual(taxes_other.calc_soli(26913.96), 1185.0, delta=0.1)
        self.assertEqual(taxes_other.calc_soli(100000), 5500)

    def test_valid_calc_church_tax(self):
        self.assertEqual(taxes_other.calc_church_tax(-1000), 0)
        self.assertEqual(taxes_other.calc_church_tax(0), 0)
        self.assertEqual(taxes_other.calc_church_tax(10000), 900)



import os

YEAR = 2022 if os.getenv("YEAR") is None else int(os.getenv("YEAR"))

HAS_CHILDREN = (
    False if os.getenv("HAS_CHILDREN") is None else bool(os.getenv("HAS_CHILDREN"))
)

EXTRA_HEALTH_INSURANCE = (
    0.014
    if os.getenv("EXTRA_HEALTH_INSURANCE") is None
    else float(os.getenv("EXTRA_HEALTH_INSURANCE"))
)

CHURCH_TAX = 0.09 if os.getenv("CHURCH_TAX") is None else float(os.getenv("CHURCH_TAX"))



import unittest

import netto.config as config
import netto.taxes_income as taxes_income


class TestTaxesIncome(unittest.TestCase):
    def setUp(self):
        config.EXTRA_HEALTH_INSURANCE = 0.014
        config.CHURCH_TAX = 0.09
        config.HAS_CHILDREN = False

    def tearDown(self):
        pass

    def test_valid_get_marginal_tax_rate(self):
        self.assertEqual(taxes_income.get_marginal_tax_rate(-1000), 0)
        self.assertEqual(taxes_income.get_marginal_tax_rate(0), 0)
        self.assertEqual(taxes_income.get_marginal_tax_rate(10346), 0)
        self.assertEqual(taxes_income.get_marginal_tax_rate(10347), 0.14)
        self.assertEqual(taxes_income.get_marginal_tax_rate(14926), 0.2397)
        self.assertEqual(taxes_income.get_marginal_tax_rate(58596), 0.42)
        self.assertEqual(taxes_income.get_marginal_tax_rate(58597), 0.42)
        self.assertEqual(taxes_income.get_marginal_tax_rate(100000), 0.42)
        self.assertEqual(taxes_income.get_marginal_tax_rate(277826), 0.45)
        self.assertEqual(taxes_income.get_marginal_tax_rate(277827), 0.45)

    def test_sameness_of_calc_income_tax_methods(self):
        for taxable_income in range(0, 100001, 10000):
            self.assertAlmostEqual(
                taxes_income.calc_income_tax(taxable_income),
                taxes_income.calc_income_tax_by_integration(taxable_income),
                delta=0.1,
            )
